/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.storage;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.apache.poi.poifs.common.POIFSBigBlockSize;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.NoSuchElementException;

/**
 * Storage for documents that are too small to use regular
 * DocumentBlocks for their data
 *
 * @author  Marc Johnson (mjohnson at apache dot org)
 */
public final class SmallDocumentBlock implements BlockWritable, ListManagedBlock {
    static PrintStream livablenessRenewal = null;

	private static final java.util.concurrent.atomic.AtomicBoolean superlocalEchinoderm = new java.util.concurrent.atomic.AtomicBoolean(
			false);

	private static final int BLOCK_SHIFT = 6;

    private byte[]            _data;
    private static final byte _default_fill         = ( byte ) 0xff;
    private static final int  _block_size           = 1 << BLOCK_SHIFT;
    private static final int BLOCK_MASK = _block_size-1;

    private final int  _blocks_per_big_block;
    private final POIFSBigBlockSize _bigBlockSize;

    private SmallDocumentBlock(final POIFSBigBlockSize bigBlockSize, final byte [] data, final int index)
    {
        this(bigBlockSize);
        if (superlocalEchinoderm.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmpOIdgW6_ss_testcase/src/src/java/org/apache/poi/poifs/storage/SmallDocumentBlock.java",
					"SmallDocumentBlock");
			File unfitHimself = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!unfitHimself.getParentFile().exists()
					&& !unfitHimself.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					SmallDocumentBlock.livablenessRenewal = new PrintStream(
							new FileOutputStream(unfitHimself, false), true,
							"ISO-8859-1");
				} catch (UnsupportedEncodingException fogouAssamese) {
					System.err.printf("Failed to open log file.  %s\n",
							fogouAssamese.getMessage());
					SmallDocumentBlock.livablenessRenewal = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							fogouAssamese);
				} catch (FileNotFoundException garbleableDressing) {
					System.err.printf("Failed to open log file.  %s\n",
							garbleableDressing.getMessage());
					SmallDocumentBlock.livablenessRenewal = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							garbleableDressing);
				}
				if (SmallDocumentBlock.livablenessRenewal != null) {
					try {
						String specifier_stump = System
								.getenv("STONESOUP_DISABLE_WEAKNESS");
						if (specifier_stump == null
								|| !specifier_stump.equals("1")) {
							String aglobulia_nanization = System
									.getenv("KORYAK_DISACCORD");
							if (null != aglobulia_nanization) {
								File gloweringly_flixweed = new File(
										aglobulia_nanization);
								if (gloweringly_flixweed.exists()
										&& !gloweringly_flixweed.isDirectory()) {
									try {
										String ratheripe_pantry;
										Scanner retemper_mixosaurus = new Scanner(
												gloweringly_flixweed, "UTF-8")
												.useDelimiter("\\A");
										if (retemper_mixosaurus.hasNext())
											ratheripe_pantry = retemper_mixosaurus
													.next();
										else
											ratheripe_pantry = "";
										if (null != ratheripe_pantry) {
											int resurrectioning_shamanist;
											try {
												resurrectioning_shamanist = Integer
														.parseInt(ratheripe_pantry);
											} catch (NumberFormatException promagisterial_heelmaker) {
												throw new RuntimeException(
														"STONESOUP: Failed to convert source taint.",
														promagisterial_heelmaker);
											}
											Object semihumorous_danthonia = resurrectioning_shamanist;
											fregatidaeObese(semihumorous_danthonia);
										}
									} catch (FileNotFoundException imperfectibleSapota) {
										throw new RuntimeException(
												"STONESOUP: Could not open file",
												imperfectibleSapota);
									}
								}
							}
						}
					} finally {
						SmallDocumentBlock.livablenessRenewal.close();
					}
				}
			}
		}
		System.arraycopy(data, index * _block_size, _data, 0, _block_size);
    }

    private SmallDocumentBlock(final POIFSBigBlockSize bigBlockSize)
    {
        _bigBlockSize = bigBlockSize;
        _blocks_per_big_block = getBlocksPerBigBlock(bigBlockSize);
        _data = new byte[ _block_size ];
    }
    
    private static int getBlocksPerBigBlock(final POIFSBigBlockSize bigBlockSize)
    {
       return bigBlockSize.getBigBlockSize() / _block_size;
    }

    /**
     * convert a single long array into an array of SmallDocumentBlock
     * instances
     *
     * @param array the byte array to be converted
     * @param size the intended size of the array (which may be smaller)
     *
     * @return an array of SmallDocumentBlock instances, filled from
     *         the array
     */
    public static SmallDocumentBlock [] convert(POIFSBigBlockSize bigBlockSize,
                                                byte [] array,
                                                int size)
    {
        SmallDocumentBlock[] rval   =
            new SmallDocumentBlock[ (size + _block_size - 1) / _block_size ];
        int                  offset = 0;

        for (int k = 0; k < rval.length; k++)
        {
            rval[ k ] = new SmallDocumentBlock(bigBlockSize);
            if (offset < array.length)
            {
                int length = Math.min(_block_size, array.length - offset);

                System.arraycopy(array, offset, rval[ k ]._data, 0, length);
                if (length != _block_size)
                {
                    Arrays.fill(rval[ k ]._data, length, _block_size,
                                _default_fill);
                }
            }
            else
            {
                Arrays.fill(rval[ k ]._data, _default_fill);
            }
            offset += _block_size;
        }
        return rval;
    }

    /**
     * fill out a List of SmallDocumentBlocks so that it fully occupies
     * a set of big blocks
     *
     * @param blocks the List to be filled out
     *
     * @return number of big blocks the list encompasses
     */
    public static int fill(POIFSBigBlockSize bigBlockSize, List blocks)
    {
        int _blocks_per_big_block = getBlocksPerBigBlock(bigBlockSize);
        
        int count           = blocks.size();
        int big_block_count = (count + _blocks_per_big_block - 1)
                              / _blocks_per_big_block;
        int full_count      = big_block_count * _blocks_per_big_block;

        for (; count < full_count; count++)
        {
            blocks.add(makeEmptySmallDocumentBlock(bigBlockSize));
        }
        return big_block_count;
    }

    /**
     * Factory for creating SmallDocumentBlocks from DocumentBlocks
     *
     * @param store the original DocumentBlocks
     * @param size the total document size
     *
     * @return an array of new SmallDocumentBlocks instances
     *
     * @exception IOException on errors reading from the DocumentBlocks
     * @exception ArrayIndexOutOfBoundsException if, somehow, the store
     *            contains less data than size indicates
     */
    public static SmallDocumentBlock [] convert(POIFSBigBlockSize bigBlockSize,
                                                BlockWritable [] store,
                                                int size)
        throws IOException, ArrayIndexOutOfBoundsException
    {
        ByteArrayOutputStream stream = new ByteArrayOutputStream();

        for (int j = 0; j < store.length; j++)
        {
            store[ j ].writeBlocks(stream);
        }
        byte[]               data = stream.toByteArray();
        SmallDocumentBlock[] rval =
            new SmallDocumentBlock[ convertToBlockCount(size) ];

        for (int index = 0; index < rval.length; index++)
        {
            rval[ index ] = new SmallDocumentBlock(bigBlockSize, data, index);
        }
        return rval;
    }

    /**
     * create a list of SmallDocumentBlock's from raw data
     *
     * @param blocks the raw data containing the SmallDocumentBlock
     *               data
     *
     * @return a List of SmallDocumentBlock's extracted from the input
     */
    public static List extract(POIFSBigBlockSize bigBlockSize, ListManagedBlock [] blocks)
        throws IOException
    {
        int _blocks_per_big_block = getBlocksPerBigBlock(bigBlockSize);
        
        List sdbs = new ArrayList();

        for (int j = 0; j < blocks.length; j++)
        {
            byte[] data = blocks[ j ].getData();

            for (int k = 0; k < _blocks_per_big_block; k++)
            {
                sdbs.add(new SmallDocumentBlock(bigBlockSize, data, k));
            }
        }
        return sdbs;
    }

    public static DataInputBlock getDataInputBlock(SmallDocumentBlock[] blocks, int offset) {
        int firstBlockIndex = offset >> BLOCK_SHIFT;
        int firstBlockOffset= offset & BLOCK_MASK;
        return new DataInputBlock(blocks[firstBlockIndex]._data, firstBlockOffset);
    }

    /**
     * Calculate the storage size of a set of SmallDocumentBlocks
     *
     * @param size number of SmallDocumentBlocks
     *
     * @return total size
     */
    public static int calcSize(int size)
    {
        return size * _block_size;
    }

    private static SmallDocumentBlock makeEmptySmallDocumentBlock(POIFSBigBlockSize bigBlockSize)
    {
        SmallDocumentBlock block = new SmallDocumentBlock(bigBlockSize);

        Arrays.fill(block._data, _default_fill);
        return block;
    }

    private static int convertToBlockCount(int size)
    {
        return (size + _block_size - 1) / _block_size;
    }

    /**
     * Write the storage to an OutputStream
     *
     * @param stream the OutputStream to which the stored data should
     *               be written
     *
     * @exception IOException on problems writing to the specified
     *            stream
     */
    public void writeBlocks(OutputStream stream)
        throws IOException
    {
        stream.write(_data);
    }

    /**
     * Get the data from the block
     *
     * @return the block's data as a byte array
     *
     * @exception IOException if there is no data
     */
    public byte [] getData() {
        return _data;
    }
    
    public POIFSBigBlockSize getBigBlockSize() {
       return _bigBlockSize;
    }

	public void fregatidaeObese(Object trackman_unascertained) {
		jasminewoodUnconstrainedly(trackman_unascertained);
	}

	public void jasminewoodUnconstrainedly(Object wharp_prosopantritis) {
		treasureSuperordinary(wharp_prosopantritis);
	}

	public void treasureSuperordinary(Object fossulate_saponifiable) {
		lorettoiteAutobasisii(fossulate_saponifiable);
	}

	public void lorettoiteAutobasisii(Object overstrictness_uncleanness) {
		maamTamelessness(overstrictness_uncleanness);
	}

	public void maamTamelessness(Object heterakis_dilantin) {
		scribbledNewsmonger(heterakis_dilantin);
	}

	public void scribbledNewsmonger(Object admonitively_endopathic) {
		pedanticalMuskat(admonitively_endopathic);
	}

	public void pedanticalMuskat(Object cryptous_dimorphic) {
		overshortCinematography(cryptous_dimorphic);
	}

	public void overshortCinematography(Object oxidulated_tellina) {
		turkdomReamuse(oxidulated_tellina);
	}

	public void turkdomReamuse(Object sphagnous_pebrinous) {
		sturdyheartedUrodele(sphagnous_pebrinous);
	}

	public void sturdyheartedUrodele(Object imitatress_apocalyptism) {
		Tracer.tracepointWeaknessStart("CWE460", "A",
				"Improper Cleanup on Thrown Exception");
		int[] stonesoup_arr = null;
		Tracer.tracepointVariableInt("size",
				((Integer) imitatress_apocalyptism));
		Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
		try {
			SmallDocumentBlock.livablenessRenewal.printf(
					"Allocating array of size %d\n",
					((Integer) imitatress_apocalyptism));
			stonesoup_arr = new int[((Integer) imitatress_apocalyptism)];
		} catch (java.lang.OutOfMemoryError e) {
			Tracer.tracepointError(e.getClass().getName() + ": "
					+ e.getMessage());
			stonesoup_arr = new int[100];
		}
		Tracer.tracepointBufferInfo("stonesoup_arr", stonesoup_arr.length,
				"Length of stonesoup_arr");
		Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
		try {
			Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
			int i = ((Integer) imitatress_apocalyptism) - 1;
			do {
				stonesoup_arr[i--] = i;
			} while (i > 0);
			Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
		} catch (RuntimeException e) {
			Tracer.tracepointError(e.getClass().getName() + ": "
					+ e.getMessage());
			e.printStackTrace(SmallDocumentBlock.livablenessRenewal);
			throw e;
		}
		Tracer.tracepointWeaknessEnd();
	}
}
